import time
from typing import TypedDict

from pulp import LpSolver
from muoblpbindings import equal_shares_utils

from muoblp.model.multi_objective_lp import MultiObjectiveLpProblem

from muoblpsolvers.common import (
    prepare_mes_parameters,
    set_selected_candidates,
)
from muoblpsolvers.mes_constrains.utils import (
    get_infeasible_constraints,
    get_feasibility_ratio,
)


class SolverOptions(TypedDict):
    cost_modification_base: float
    max_iterations: int


class MethodOfEqualSharesConstrainsSolver(LpSolver):
    """
    Info:
        Method Of Equal Shares with Constraints solver
    """

    def __init__(self, solver_options):
        super().__init__()
        self.solver_options: SolverOptions = solver_options

    def actualSolve(self, lp: MultiObjectiveLpProblem):
        print(
            f"Starting MethodOfEqualSharesConstrainsSolver {self.solver_options}"
        )
        """
        Parameters:
            lp: Instance of MultiObjectiveLpProblem
        """
        (
            projects,
            costs,
            voters,
            approvals_utilities,
            total_utilities,
            total_budget,
        ) = prepare_mes_parameters(lp)

        iteration = 0
        while iteration < self.solver_options["max_iterations"]:
            # Run MES
            start_time = time.time()
            selected = equal_shares_utils(
                voters,
                projects,
                costs,
                approvals_utilities,
                total_utilities,
                total_budget,
            )
            print(f"FINISHED MES {time.time() - start_time:.2f} s\n")
            set_selected_candidates(lp, selected)

            # Check constraints
            infeasible = get_infeasible_constraints(lp)
            for constraint in infeasible:
                print(
                    f"FEAS_RATIO|{iteration}|{constraint.name}|{get_feasibility_ratio(constraint):.6f}"
                )

            if len(infeasible) == 0:
                print(
                    "============== all constraints fulfilled =============="
                )
                break

            # TODO: Extract to parametrized strategy
            # Modify prices
            for constraint in infeasible:
                feasibility_ratio = get_feasibility_ratio(
                    constraint
                )  # ratio: [0, inf)
                cost_modification_ratio = feasibility_ratio * (
                    self.solver_options["cost_modification_base"] ** iteration
                )  # exponential backoff
                affected_candidates = [
                    candidate.name for candidate in constraint.keys()
                ]
                print(
                    f"Modifying cost of {len(affected_candidates)} variables with ratio {cost_modification_ratio:.4f}"
                )
                for candidate in affected_candidates:
                    costs[candidate] = int(
                        costs[candidate] * cost_modification_ratio
                    )

            iteration += 1



from pulp import LpSolver, lpSum, PULP_CBC_CMD, GUROBI_CMD

from muoblp.model.multi_objective_lp import MultiObjectiveLpProblem


class SummedObjectivesLpSolver(LpSolver):
    """

    Info:
        Example dummy solver that sums multiple objectives.
        Parameter flag to use gurobi solver instead of default PULP one.
    """

    def __init__(self, use_gurobi: bool = False):
        super().__init__()
        self.use_gurobi = use_gurobi

    def actualSolve(self, lp: MultiObjectiveLpProblem):
        """
        Parameters:
            lp: Instance of MultiObjectiveLpProblem
        """
        lp.setObjective(lpSum(lp.objectives))
        solver_cmd = (
            GUROBI_CMD() if self.use_gurobi else PULP_CBC_CMD(msg=False)
        )
        return solver_cmd.actualSolve(lp)



from muoblp.model.multi_objective_lp import MultiObjectiveLpProblem

from muoblpsolvers.types import CandidateId, VoterId


def break_ties(
    cost: dict[CandidateId, float],
    total_utility: dict[CandidateId, int],
    choices: list[CandidateId],
) -> CandidateId:
    remaining = choices.copy()
    best_cost = min(cost[c] for c in remaining)
    remaining = [c for c in remaining if cost[c] == best_cost]
    best_count = max(total_utility[c] for c in remaining)
    remaining = [c for c in remaining if total_utility[c] == best_count]
    if len(remaining) > 1:
        print(
            "Tie-breakign failed: tie between projects {remaining}. Selecting first one."
        )
        # raise Exception(
        #     f"Tie-breaking failed: tie between projects {remaining} could not be resolved. Another tie-breaking needs to be added."
        # )
    return remaining[0]


def equal_shares_exponential(
    voters: list[VoterId],
    projects: list[CandidateId],
    cost: dict[CandidateId, float],
    approvals_utilities: dict[CandidateId, list[tuple[VoterId, int]]],
    total_utility: dict[CandidateId, int],
    lp: MultiObjectiveLpProblem,
    budget_init: float,
):
    budget: dict[VoterId, float] = {voter: budget_init for voter in voters}
    remaining = {}  # remaining candidate -> previous effective vote count Dict[CandidateId, int]
    for candidate in projects:
        if cost[candidate] > 0 and len(approvals_utilities[candidate]) > 0:
            remaining[candidate] = total_utility[candidate]
    winners = []
    i = 0
    while len(remaining) and max(remaining.values()) > 0:
        i += 1
        budget_init *= 2
        try:
            for voter in voters:
                budget[voter] += budget_init
        except OverflowError:
            print(i)
            break

        while len(remaining):
            best: list[CandidateId] = []
            best_eff_vote_count = 0
            # go through remaining candidates in order of decreasing previous effective vote count
            remaining_sorted = sorted(
                remaining, key=lambda c: remaining[c], reverse=True
            )
            for candidate in remaining_sorted:
                previous_eff_vote_count = remaining[candidate]
                if previous_eff_vote_count < best_eff_vote_count:
                    # c cannot be better than the best so far
                    break
                money_behind_now = sum(
                    budget[voter]
                    for voter, _ in approvals_utilities[candidate]
                )
                if money_behind_now < cost[candidate]:
                    # c is not affordable in this round - try again after increasing budgets
                    # del remaining[c]
                    continue

                # calculate the effective vote count of c
                approvals_utilities[candidate].sort(
                    key=lambda voter_utility: budget[voter_utility[0]]
                    / voter_utility[1]
                )
                paid_so_far = 0
                denominator = remaining[candidate]  # total utility of c
                for voter, utility in approvals_utilities[candidate]:
                    # compute payment if remaining approvers pay proportional to their utility
                    max_payment = (cost[candidate] - paid_so_far) / denominator
                    eff_vote_count = cost[candidate] / max_payment
                    if max_payment * utility > budget[voter]:
                        # voter cannot afford the payment, so pays entire remaining budget
                        paid_so_far += budget[voter]
                        denominator -= utility
                    else:
                        # i (and all later approvers) can afford the payment; stop here
                        # TODO: make sure not to drop eff_vote_count below zero
                        remaining[candidate] = eff_vote_count
                        if eff_vote_count > best_eff_vote_count:
                            best_eff_vote_count = eff_vote_count
                            best = [candidate]
                        elif eff_vote_count == best_eff_vote_count:
                            best.append(candidate)
                        break
            if not best:
                # no remaining candidates are affordable
                break
            selected = break_ties(cost, total_utility, best)
            ### Try selecting best
            candidate_variable = [
                variable
                for variable in lp.variables()
                if variable.name == selected
            ][0]
            candidate_variable.setInitialValue(1)

            ### remove if selecting c breaks feasibility
            if not lp.valid():
                candidate_variable.setInitialValue(0)
                del remaining[selected]
                continue

            winners.append(selected)
            del remaining[selected]
            # charge the approvers of selected
            best_max_payment = cost[selected] / best_eff_vote_count
            for voter, utility in approvals_utilities[selected]:
                payment = best_max_payment * utility
                budget[voter] -= min(payment, budget[voter])

    return winners



from typing import List

from pulp import LpConstraint, LpConstraintGE, LpConstraintLE

from muoblp.model.multi_objective_lp import MultiObjectiveLpProblem


def get_feasibility_ratio(constraint: LpConstraint) -> float:
    """
    :rtype: object
    """
    # ratio: [0, inf)
    value = constraint.value()
    target = constraint.constant
    return (value - target) / abs(target)


# def get_modification_ratio(feasibility_ratio: float, lower: float, upper: float) -> float:
#     return lower + (upper - lower) * feasibility_ratio


def get_infeasible_constraints(
    problem: MultiObjectiveLpProblem,
) -> List[LpConstraint]:
    return [
        constraint
        for constraint in problem.constraints.values()
        if (constraint.sense == LpConstraintGE and constraint.value() < 0)
        or (constraint.sense == LpConstraintLE and constraint.value() > 0)
    ]



import time

from pulp import LpSolver

from muoblp.model.multi_objective_lp import MultiObjectiveLpProblem
from muoblpbindings import equal_shares_utils

from muoblpsolvers.common import (
    prepare_mes_parameters,
    set_selected_candidates,
)


class MethodOfEqualSharesUtilitySolver(LpSolver):
    """
    Info:
        Solver that executes Methods of Equal Shares to find solution
    """

    def __init__(self):
        super().__init__()

    def actualSolve(self, lp: MultiObjectiveLpProblem):
        (
            projects,
            costs,
            voters,
            approvals_utilities,
            total_utilities,
            total_budget,
        ) = prepare_mes_parameters(lp)

        start_time = time.time()
        print(f"STARTING MES UTILS {start_time}")
        selected = equal_shares_utils(
            voters,
            projects,
            costs,
            approvals_utilities,
            total_utilities,
            total_budget,
        )
        print(f"FINISHED MES UTILS {time.time() - start_time}")

        set_selected_candidates(lp, selected)
        return



from typing import TypedDict

from pulp import LpSolver

from muoblp.model.multi_objective_lp import MultiObjectiveLpProblem

from muoblpsolvers.common import (
    prepare_mes_parameters,
)
from muoblpsolvers.mes_exponential.mes_exponential import (
    equal_shares_exponential,
)


class SolverOptions(TypedDict):
    budget_init: int


class MethodOfEqualSharesExponentialSolver(LpSolver):
    """
    Info:
        Method Of Equal Shares Exponential variant solver
    """

    def __init__(self, solver_options):
        super().__init__()
        self.solver_options: SolverOptions = solver_options

    def actualSolve(self, lp: MultiObjectiveLpProblem):
        print(
            f"Starting MethodOfEqualSharesExponentialSolver {self.solver_options}"
        )
        """
        Parameters:
            lp: Instance of MultiObjectiveLpProblem
        """
        (
            projects,
            costs,
            voters,
            approvals_utilities,
            total_utilities,
            total_budget,
        ) = prepare_mes_parameters(lp)

        equal_shares_exponential(
            voters,
            projects,
            costs,
            approvals_utilities,
            total_utilities,
            lp,
            self.solver_options["budget_init"],
        )



from collections import defaultdict

from muoblp.model.multi_objective_lp import MultiObjectiveLpProblem
from pulp import LpConstraintLE, LpConstraint

from muoblpsolvers.types import (
    CandidateId,
    Cost,
    TotalBudget,
    Utility,
    VoterId,
)


def get_total_budget_constraint(lp: MultiObjectiveLpProblem) -> LpConstraint:
    all_candidates: set[CandidateId] = set(
        [
            variable.name
            for variable in lp.variables()
            if variable.name != "__dummy"
        ]
    )

    pb_constraints = []
    for constraint in lp.constraints.values():
        candidates = set([variable.name for variable, _ in constraint.items()])
        if candidates == all_candidates and constraint.sense == LpConstraintLE:
            pb_constraints.append(constraint)

    if len(pb_constraints) == 0:
        raise Exception("Problem does not have PB constraint")
    if len(pb_constraints) > 1:
        pass  # TODO - fix duplicated constraints for single district and city wide
        # raise Exception("Problem has too many PB constraint")
    return pb_constraints[0]


def prepare_mes_parameters(
    lp: MultiObjectiveLpProblem,
) -> tuple[
    list[CandidateId],
    dict[CandidateId, Cost],
    list[VoterId],
    dict[CandidateId, list[tuple[VoterId, Utility]]],
    dict[CandidateId, Utility],
    TotalBudget,
]:
    projects: list[CandidateId] = [
        candidate.name
        for candidate in lp.variables()
        if candidate.name != "__dummy"
    ]

    costs: dict[CandidateId, Cost] = {
        candidate.name: cost
        # TODO: some constraints have repeated variables, but we just override them with the same value
        for constraint in lp.constraints.values()
        # [C_ub_Bieńczyce: 25000 V_BO.D16.10_24 + 40500 V_BO.D16.11_24 <= 100000, ...]
        for candidate, cost in constraint.items()
    }

    voters: list[VoterId] = [voter.name for voter in lp.objectives]

    approvals_utilities: dict[CandidateId, list[tuple[VoterId, Utility]]] = (
        defaultdict(list)
    )
    for voter in (
        lp.objectives
    ):  # [T_6080: 80550 V_BO.D10.14_24 + 340000 V_BO.D10.1_24, ....]
        for candidate, utility in voter.items():
            approvals_utilities[candidate.name] += [(voter.name, utility)]

    total_utilities: dict[CandidateId, Utility] = {
        candidate: sum([u for v, u in voters_utilities])
        for candidate, voters_utilities in approvals_utilities.items()
    }

    no_vote_projects = []
    for project in projects:
        if project not in approvals_utilities:
            no_vote_projects.append(project)

    for project in no_vote_projects:
        print(f"Removing project with zero votes {project}")
        projects.remove(project)
        del costs[project]

    total_budget = abs(get_total_budget_constraint(lp).constant)
    return (
        projects,
        costs,
        voters,
        approvals_utilities,
        total_utilities,
        total_budget,
    )


def set_selected_candidates(
    problem: MultiObjectiveLpProblem, selected: list[str]
) -> None:
    for variable in problem.variables():
        variable.setInitialValue(1 if variable.name in selected else 0)
